import os
import base64
import logging
import pickle
import sys

from datetime import timedelta

from django.conf import settings
from django.core.management import call_command
from django.core.exceptions import ImproperlyConfigured
from django.db import transaction
from django.utils.timezone import now

try:
    from celery import Task
    from celery import shared_task
    from celery.exceptions import CeleryError
    from kombu.utils.uuid import uuid
except ImportError:
    raise ImproperlyConfigured('Missing celery library, please install it')

from .models import CeleryTaskLog, CeleryTaskLogState
from .utils import LogStringIO


LOGGER = logging.getLogger(__name__)


class LoggedTask(Task):

    abstract = True
    logger_level = logging.WARNING
    retry_error_message = (
        'Task "{task_name}" ({task}) failed on exception: "{exception}", attempt: "{attempt}" and will be retried'
    )
    fail_error_message = 'Task "{task_name}" ({task}) failed on exception: "{exception}", attempt: "{attempt}"'
    stale_time_limit = None
    # Support set retry delay in list. Retry countdown value is get from list where index is attempt
    # number (request.retries)
    default_retry_delays = None

    @property
    def task_log(self):
        return CeleryTaskLog.objects.get(
            celery_task_id=str(self.request.id),
            retries=self.request.retries
        )

    def __call__(self, *args, **kwargs):
        """
        Overrides parent which works with thread stack. We didn't want to allow change context which was generated in
        one of apply methods. Call task directly is now disallowed.
        """
        req = self.request_stack.top

        if not req or req.called_directly:
            raise CeleryError(
                'Task cannot be called directly. Please use apply, apply_async or apply_async_on_commit methods'
            )

        if req._protected:
            raise CeleryError('Request is protected')
        # request is protected (no usage in celery but get from function _install_stack_protection in
        # celery library)
        req._protected = 1

        # Every set attr is sent here
        self.request.output_stream = LogStringIO()
        self.on_start_task(self.task_log, args, kwargs)
        return self.run(*args, **kwargs)

    def on_apply_task(self, task_log, args, kwargs, options):
        """
        On apply task is invoked before task was prepared. Therefore task request context is not prepared.
        :param task_log: logged celery task instance
        :param args: task args
        :param kwargs: task kwargs
        :param options: input task options
        """

    def on_retry_task(self, task_log, args, kwargs, exc):
        """
        On retry task is invoked before task was retried.
        :param task_log: logged celery task instance
        :param args: task args
        :param kwargs: task kwargs
        :param exc: raised exception which caused retry
        """

        LOGGER.log(self.logger_level, self.retry_error_message.format(
            attempt=self.request.retries, exception=str(exc), task=task_log, task_name=task_log.name, **(kwargs or {})
        ))
        task_log.change_and_save(
            state=CeleryTaskLogState.RETRIED,
            error_message=str(exc),
            output=self.request.output_stream.getvalue(),
            retries=self.request.retries
        )

    def on_start_task(self, task_log, args, kwargs):
        """
        On start task is invoked during task started.
        :param task_log: logged celery task instance
        :param args: task args
        :param kwargs: task kwargs
        """

        task_log.change_and_save(
            state=CeleryTaskLogState.ACTIVE,
            start=now()
        )

    def on_success_task(self, task_log, args, kwargs, retval):
        """
        On start task is invoked if task is successful.
        :param task_log: logged celery task instance
        :param args: task args
        :param kwargs: task kwargs
        :param retval: return value of a task
        """
        if retval:
            self.request.output_stream.write('Return value is "{}"'.format(retval))

        task_log.change_and_save(
            state=CeleryTaskLogState.SUCCEEDED,
            stop=now(),
            output=self.request.output_stream.getvalue(),
            retries=self.request.retries
        )

    def on_success(self, retval, task_id, args, kwargs):
        self.on_success_task(self.task_log, args, kwargs, retval)

    def on_failure_task(self, task_log, args, kwargs, exc):
        """
        On failure task is invoked if task end with some exception.
        :param task_log: logged celery task instance
        :param args: task args
        :param kwargs: task kwargs
        :param exc: raised exception which caused retry
        """

        LOGGER.log(self.logger_level, self.fail_error_message.format(
            attempt=self.request.retries, exception=str(exc), task=task_log, task_name=task_log.name, **kwargs
        ))
        task_log.change_and_save(
            state=CeleryTaskLogState.FAILED,
            stop=now(),
            error_message=str(exc),
            output=self.request.output_stream.getvalue(),
            retries=self.request.retries
        )

    def expire_task(self, task_log):
        task_log.change_and_save(
            state=CeleryTaskLogState.EXPIRED,
            stop=now(),
            error_message='Task execution was expired by command'
        )

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        try:
            self.on_failure_task(self.task_log, args, kwargs, exc)
        except CeleryTaskLog.DoesNotExist:
            pass

    def _get_eta(self, options, now):
        if options.get('countdown') is not None:
            return now + timedelta(seconds=options['countdown'])
        elif options.get('eta'):
            return options['eta']
        else:
            return now

    def _get_time_limit(self, options):
        if options.get('time_limit') is not None:
            return options['time_limit']
        elif self.soft_time_limit is not None:
            return self.soft_time_limit
        else:
            return self._get_app().conf.task_time_limit

    def _get_stale_time_limit(self, stale_time_limit):
        if stale_time_limit is not None:
            return stale_time_limit
        elif self.stale_time_limit is not None:
            return self.stale_time_limit
        elif hasattr(settings, 'CELERYD_TASK_STALE_TIME_LIMIT'):
            return settings.CELERYD_TASK_STALE_TIME_LIMIT
        else:
            return None

    def _get_expires(self, options, now, stale_time_limit):
        if options.get('expires') is not None:
            return options['expires']
        elif self.expires is not None:
            return self.expires
        elif self._get_stale_time_limit(stale_time_limit) is not None and self._get_time_limit(options) is not None:
            return now + timedelta(
                seconds=(self._get_stale_time_limit(stale_time_limit)) - self._get_time_limit(options)
            )
        else:
            return None

    def _create_task(self, options, task_args, task_kwargs, celery_task_id, now, stale_time_limit):
        task_input = []
        if task_args:
            task_input += [str(v) for v in task_args]
        if task_kwargs:
            task_input += ['{}={}'.format(k, v) for k, v in task_kwargs.items()]

        return CeleryTaskLog.objects.create(
            celery_task_id=celery_task_id,
            name=self.name,
            state=CeleryTaskLogState.WAITING,
            queue_name=options.get('queue', getattr(self, 'queue', settings.CELERY_DEFAULT_QUEUE)),
            input=', '.join(task_input),
            task_args=task_args,
            task_kwargs=task_kwargs,
            estimated_time_of_arrival=options['eta'],
            expires=options['expires'],
            stale=(
                now + timedelta(seconds=self._get_stale_time_limit(stale_time_limit))
                if self._get_stale_time_limit(stale_time_limit) is not None else None
            ),
            retries=options.get('retries', 0)
        )

    def _update_options(self, options, now, stale_time_limit):
        options['eta'] = self._get_eta(options, now)
        options['expires'] = self._get_expires(options, now, stale_time_limit)
        options.pop('countdown', None)
        return options

    def apply_async_on_commit(self, args=None, kwargs=None, **options):
        if sys.argv[1:2] == ['test']:
            self.apply_async(args=args, kwargs=kwargs, **options)
        else:
            self_inst = self
            transaction.on_commit(
                lambda: self_inst.apply_async(args=args, kwargs=kwargs, **options)
            )

    def apply(self, args=None, kwargs=None, task_id=None, stale_time_limit=None, **options):
        apply_time = now()
        options = self._update_options(options, apply_time, stale_time_limit)
        celery_task_id = task_id or uuid()
        task_log = self._create_task(options, args, kwargs, celery_task_id, apply_time, stale_time_limit)
        self.on_apply_task(task_log, args, kwargs, options)
        return super().apply(args=args, kwargs=kwargs, task_id=celery_task_id, **options)

    def apply_async(self, args=None, kwargs=None, task_id=None, stale_time_limit=None, **options):
        celery_task_id = task_id or uuid()
        app = self._get_app()
        if app.conf.task_always_eager:
            # Is called apply method which prepare task itself
            return super().apply_async(args=args, kwargs=kwargs, task_id=celery_task_id,
                                       stale_time_limit=stale_time_limit, **options)
        else:
            apply_time = now()
            options = self._update_options(options, apply_time, stale_time_limit)
            task_log = self._create_task(options, args, kwargs, celery_task_id, apply_time, stale_time_limit)
            self.on_apply_task(task_log, args, kwargs, options)
            return super().apply_async(args=args, kwargs=kwargs, task_id=celery_task_id, **options)

    def retry(self, args=None, kwargs=None, exc=None, throw=True,
              eta=None, countdown=None, max_retries=None, default_retry_delays=None, **options):

        self.on_retry_task(self.task_log, args, kwargs, exc)

        # Celery retry not working in eager mode. This simple hack fix it.
        self.request.is_eager = False

        if (default_retry_delays or (
                max_retries is None and eta is None and countdown is None and max_retries is None
                and self.default_retry_delays)):
            default_retry_delays = self.default_retry_delays if default_retry_delays is None else default_retry_delays
            max_retries = len(default_retry_delays)
            countdown = default_retry_delays[self.request.retries] if self.request.retries < max_retries else None
        return super().retry(
            args=args, kwargs=kwargs, exc=exc, throw=throw,
            eta=eta, countdown=countdown, max_retries=max_retries, **options
        )


def obj_to_string(obj):
    return base64.encodebytes(pickle.dumps(obj)).decode('utf8')


def string_to_obj(obj_string):
    return pickle.loads(base64.decodebytes(obj_string.encode('utf8')))


@shared_task(
    base=LoggedTask,
    bind=True,
    name='call_django_command'
)
def call_django_command(self, command_name, command_args=None):
    command_args = [] if command_args is None else command_args
    call_command(
        command_name,
        settings=os.environ.get('DJANGO_SETTINGS_MODULE'),
        *command_args,
        stdout=self.request.output_stream,
        stderr=self.request.output_stream,
    )



# Generated by Django 2.0 on 2019-08-15 13:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('security', '0011_auto_20190815_1508'),
    ]

    operations = [
        migrations.AlterField(
            model_name='commandlog',
            name='changed_at',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at'),
        ),
        migrations.AlterField(
            model_name='commandlog',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='changed_at',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='path',
            field=models.CharField(blank=True, db_index=True, max_length=255, verbose_name='URL path'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='slug',
            field=models.CharField(blank=True, db_index=True, max_length=255, null=True, verbose_name='slug'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='changed_at',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='path',
            field=models.CharField(blank=True, db_index=True, max_length=255, verbose_name='URL path'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='slug',
            field=models.CharField(blank=True, db_index=True, max_length=255, null=True, verbose_name='slug'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-08-15 18:12
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('reversion_log', '0002_auto_20180418_0925'),
    ]

    operations = [
        migrations.AlterField(
            model_name='inputrequestrevision',
            name='logged_request',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                    related_name='input_logged_request_revisions', to='security.InputLoggedRequest',
                                    verbose_name='logged request'),
        ),
        migrations.AlterField(
            model_name='inputrequestrevision',
            name='revision',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE,
                                       related_name='input_logged_request_revision', to='reversion.Revision',
                                       verbose_name='revision'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2019-11-03 22:05
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('security', '0016_migration'),
    ]

    operations = [
        migrations.AlterField(
            model_name='celerytasklog',
            name='celery_task_id',
            field=models.CharField(max_length=250, verbose_name='celery ID'),
        ),
        migrations.AlterUniqueTogether(
            name='celerytasklog',
            unique_together=set([('celery_task_id', 'retries')]),
        ),
    ]



from django.core.management.base import CommandError

from security.models import InputLoggedRequest, OutputLoggedRequest, CommandLog, CeleryTaskLog
from security.utils import PurgeLogsBaseCommand


class Command(PurgeLogsBaseCommand):

    timestamp_field = 'created_at'

    models = {
        'input-request': InputLoggedRequest,
        'output-request': OutputLoggedRequest,
        'command': CommandLog,
        'celery': CeleryTaskLog
    }

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument('--type', action='store', dest='type', default='input',
                            help='Tells Django what type of requests should be removed '
                                 '(input-request/output-request/command/celery).')

    def handle(self, expiration, type, **options):
        self.type = type
        super().handle(expiration, **options)

    def get_model(self):
        model = self.models.get(self.type)
        if model:
            return model
        else:
            raise CommandError('Type can be only input-request, output-request, command or celery')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2019-10-30 13:15
from __future__ import unicode_literals

from django.db.models import F
from django.db.models.functions import Cast
from django.db import migrations, models
import django.utils.timezone

import jsonfield.fields


def set_estimated_time_and_celery_task_id_fields(apps, schema_editor):
    CeleryTaskLog = apps.get_model('security', 'CeleryTaskLog')
    CeleryTaskLog.objects.all().update(
        estimated_time_of_arrival=F('created_at'),
        celery_task_id=Cast('id', models.CharField(max_length=250))
    )


class Migration(migrations.Migration):

    dependencies = [
        ('security', '0015_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='celerytasklog',
            name='estimated_time_of_arrival',
            field=models.DateTimeField(
                default=django.utils.timezone.now, verbose_name='estimated time of arrival'
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='celerytasklog',
            name='retries',
            field=models.PositiveSmallIntegerField(default=0, verbose_name='retries'),
        ),
        migrations.AddField(
            model_name='celerytasklog',
            name='expires',
            field=models.DateTimeField(
                verbose_name='time of expiration', null=True, blank=True
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='celerytasklog',
            name='task_args',
            field=jsonfield.fields.JSONField(blank=True, null=True, verbose_name='task args'),
        ),
        migrations.AddField(
            model_name='celerytasklog',
            name='task_kwargs',
            field=jsonfield.fields.JSONField(blank=True, null=True, verbose_name='task kwargs'),
        ),
        migrations.AddField(
            model_name='celerytasklog',
            name='celery_task_id',
            field=models.CharField(null=True, blank=True, max_length=250, verbose_name='celery ID'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='celerytasklog',
            name='stale',
            field=models.DateTimeField(blank=True, null=True, verbose_name='stale task time'),
        ),
        migrations.RunPython(set_estimated_time_and_celery_task_id_fields),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.26 on 2019-11-26 15:29
from __future__ import unicode_literals

from django.db import migrations, models
import enumfields.fields
import security.enums


class Migration(migrations.Migration):

    dependencies = [
        ('security', '0017_migration'),
    ]

    operations = [
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='ip',
            field=models.GenericIPAddressField(db_index=True, verbose_name='IP address'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='response_timestamp',
            field=models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='response timestamp'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='type',
            field=enumfields.fields.NumEnumField(db_index=True, default=1, enum=security.enums.InputLoggedRequestType,
                                                 verbose_name='type'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='response_timestamp',
            field=models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='response timestamp'),
        ),
    ]



import atexit
import gzip
import json
import os
import re
import sys
import traceback
from datetime import datetime, time, timedelta
from importlib import import_module
from io import StringIO

from six.moves import input

from django.core import serializers
from django.core.exceptions import ImproperlyConfigured
from django.core.files.base import ContentFile
from django.core.management.base import BaseCommand
from django.core.serializers.json import DjangoJSONEncoder
from django.urls import resolve
from django.urls.exceptions import Resolver404
from django.utils import timezone
from django.utils.encoding import force_bytes, force_text
from django.utils.module_loading import import_string
from django.utils.timezone import now, utc

from .config import settings


UNIT_OPTIONS = {
    'h': lambda amount: timezone.now() - timedelta(hours=amount),
    'd': lambda amount: timezone.now() - timedelta(days=amount),
    'w': lambda amount: timezone.now() - timedelta(weeks=amount),
    'm': lambda amount: timezone.now() - timedelta(days=(30 * amount)),  # 30-day month
    'y': lambda amount: timezone.now() - timedelta(weeks=(52 * amount)),  # 364-day year
}

storage = import_string(settings.BACKUP_STORAGE_CLASS)()


def is_base_collection(v):
    return isinstance(v, (list, tuple, set))


def get_throttling_validators(name):
    try:
        return getattr(import_module(settings.DEFAULT_THROTTLING_VALIDATORS_PATH), name)
    except (ImportError, AttributeError):
        raise ImproperlyConfigured('Throttling validator configuration {} is not defined'.format(name))


def get_headers(request):
    regex = re.compile('^HTTP_')
    return dict((regex.sub('', header), value) for (header, value)
                in request.META.items() if header.startswith('HTTP_'))


def remove_nul_from_string(value):
    return value.replace('\x00', '')


class LogStringIO(StringIO):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._last_newline = 0

    def write(self, s):
        line_first = True
        for line in s.split('\n'):
            if not line_first:
                super().write('\n')
                self._last_newline = self.tell()
            cursor_first = True
            for cursor_block in line.split('\r'):
                if not cursor_first:
                    self.seek(self._last_newline)
                    self.truncate()
                cursor_first = False
                super().write(cursor_block)
            line_first = False

    def isatty(self):
        return True


class TeeStringIO:
    """
    StringIO that additionally writes to another stream(s).
    """
    ANSI_ESCAPE_REGEX = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')

    def __init__(self, out, log_stream):
        self._out = out
        self._log_stream = log_stream
        self._last_newline = 0
        super().__init__()

    def write(self, s):
        if self.isatty():
            self._out.write(s)
        else:
            self._out.write(self.ANSI_ESCAPE_REGEX.sub('', s))
        self._log_stream.write(s)

    def __getattr__(self, name):
        return getattr(self._out, name)

    def isatty(self):
        return hasattr(self._out, 'isatty') and self._out.isatty()

    def flush(self):
        self._out.flush()
        self._log_stream.flush()


class CommandLogger(StringIO):
    """
    A helper class that runs a django command and logs details about its run into DB.
    """

    def __init__(self, command_function, command_args=None, command_kwargs=None, stdout=None, stderr=None, **kwargs):
        """
        Initializes the command logger.

        Arguments:
            command_function: Callable that implements the command.
            stdout: Custom stream where command's standard output will be written.
            stderr: Custom stream where command's error output will be written.
            **kwargs: Keyword arguments passed to CommandLog model.
        """
        assert 'name' in kwargs, 'Key name must be provided in kwargs'

        self.command_function = command_function
        self.command_args = command_args or ()
        self.command_kwargs = command_kwargs or {}

        self.kwargs = kwargs
        self.stdout = stdout if stdout else sys.stdout
        self.stderr = stderr if stderr else sys.stderr

    def run(self):
        """
        Runs the command function and returns its return value or re-raises any exceptions. The run of the command will
        not be logged if it is in excluded commands setting.
        """
        from security.models import CommandLog

        if self.kwargs['name'] in settings.COMMAND_LOG_EXCLUDED_COMMANDS:
            return self.command_function(
                stdout=self.stdout, stderr=self.stderr, *self.command_args, **self.command_kwargs
            )

        self.output = LogStringIO()
        stdout = TeeStringIO(self.stdout, self.output)
        stderr = TeeStringIO(self.stderr, self.output)

        self.command_log = CommandLog.objects.create(start=now(), **self.kwargs)

        # register call of the finish method in case the command exits the interpreter prematurely
        atexit.register(lambda: self._finish(error_message='Command was killed'))

        try:
            ret_val = self.command_function(
                stdout=stdout, stderr=stderr, *self.command_args, **self.command_kwargs
            )
            self._finish(success=True)
            return ret_val
        except Exception as ex:  # pylint: disable=W0703
            self._finish(success=False, error_message=traceback.format_exc())
            raise ex

    def _finish(self, success=False, error_message=None):
        if self.command_log.stop is None:
            self.command_log.change_and_save(
                output=self.output.getvalue(),
                stop=now(),
                is_successful=success,
                error_message=error_message
            )


class PurgeLogsBaseCommand(BaseCommand):

    help = ""
    args = '[amount duration]'

    model = None
    timestamp_field = None

    def get_timestamp_field(self):
        return self.timestamp_field

    def get_model(self):
        return self.model

    def add_arguments(self, parser):
        parser.add_argument('--expiration', action='store', dest='expiration',
                            help='Sets the timedelta from which logs will be removed.', required=True)
        parser.add_argument('--noinput', action='store_false', dest='interactive', default=True,
                            help='Tells Django to NOT prompt the user for input of any kind.')
        parser.add_argument('--backup', action='store', dest='backup', default=False,
                            help='Tells Django where to backup removed logs.')

    def serialize(self, qs):
        data = serializers.serialize('python', qs)

        for obj_data in data:
            del obj_data['pk']
        return data

    def backup_logs_and_clean_data(self, qs, backup_path):
        for timestamp in qs.datetimes(self.get_timestamp_field(), 'day', tzinfo=utc):
            min_timestamp = datetime.combine(timestamp, time.min).replace(tzinfo=utc)
            max_timestamp = datetime.combine(timestamp, time.max).replace(tzinfo=utc)
            qs_filtered_by_day = qs.filter(
                **{'{}__range'.format(self.get_timestamp_field()): (min_timestamp, max_timestamp)}
            )
            if backup_path:
                log_file_path = os.path.join(backup_path, force_text(timestamp.date()))

                if storage.exists('{}.json.zip'.format(log_file_path)):
                    i = 0
                    while storage.exists('{}({}).json.zip'.format(log_file_path, i)):
                        i += 1
                    log_file_path = '{}({})'.format(log_file_path, i)

                self.stdout.write(4 * ' ' + 'generate backup file: ' + log_file_path)
                storage.save(
                    '{}.json.zip'.format(log_file_path),
                    ContentFile(gzip.compress(
                        force_bytes(json.dumps(self.serialize(qs_filtered_by_day), cls=DjangoJSONEncoder, indent=5)),
                    )),
                )
            self.stdout.write(4 * ' ' + 'clean logs for day {}'.format(timestamp.date()))
            qs_filtered_by_day.delete()

    def handle(self, expiration, **options):
        # Check we have the correct values
        unit = expiration[-1]
        amount = expiration[0:-1]
        try:
            amount = int(amount)
        except ValueError:
            raise CommandError('Invalid expiration format')

        if unit not in UNIT_OPTIONS:
            raise CommandError('Invalid expiration format')

        model = self.get_model()
        qs = model.objects.filter(**{'{}__lte'.format(self.get_timestamp_field()): UNIT_OPTIONS[unit](amount)})

        if qs.count() == 0:
            self.stdout.write('There are no logs to delete.')
        else:
            if options.get('interactive'):
                confirm = input('''
                You have requested a database reset.
                This will IRREVERSIBLY DESTROY any
                logs created before {} {}
                ago. That is a total of {} logs.
                Are you sure you want to do this?
                Type 'yes' to continue, or 'no' to cancel: '''.format(amount, unit, qs.count()))
            else:
                confirm = 'yes'

            if confirm == 'yes':
                try:
                    self.stdout.write('Clean data')
                    self.backup_logs_and_clean_data(qs, options.get('backup'))
                except IOError as ex:
                    self.stderr.write(force_text(ex))


def regex_sub_groups_global(pattern, repl, string):
    """
    Globally replace all groups inside pattern with `repl`.
    If `pattern` doesn't have groups the whole match is replaced.
    """
    for search in reversed(list(re.finditer(pattern, string))):
        for i in range(len(search.groups()), 0 if search.groups() else -1, -1):
            start, end = search.span(i)
            string = string[:start] + repl + string[end:]
    return string


def get_view_from_request_or_none(request):
    try:
        return resolve(request.path).func
    except Resolver404:
        return None
